/**
 * 
 */
package sim.math;

import java.nio.DoubleBuffer;
import java.util.Arrays;

import sim.util.SArrays;

/**
 * La classe <b>SMatrix4x4</b> reprsentant une matrice de '4' lignes et '4' colonnes. 
 * Les composantes de la matrice respectent la notation suivante :
 * <p>M[i][j] = </p> 
 * <p>| m00  m01  m02  m03 |</p>
 * <p>| m10  m11  m12  m13 |</p>
 * <p>| m20  m21  m22  m23 |</p>
 * <p>| m30  m31  m32  m33 |</p>
 *  
 * Puisque l'implmentation de cette matrice se fera  l'aide d'un tableau  une dimension, on retrouve l'indexation suivante :
 * <p>M[i][j] = </p> 
 * <p>|  0    1    2    3  |</p>
 * <p>|  4    5    6    7  |</p>
 * <p>|  8    9   10   11  |</p>
 * <p>| 12   13   14   15  |</p>
 * 
 * @author Simon Vzina
 * @since 2015-05-27
 * @version 2020-07-21
 */
public class SMatrix4x4 {

  /**
   * La constance <b>EMPTY</b> reprsente le tableau d'une matrice vide.
   */
  private static final double[] EMPTY = { 0.0, 0.0, 0.0, 0.0,
                                          0.0, 0.0, 0.0, 0.0,
                                          0.0, 0.0, 0.0, 0.0,
                                          0.0, 0.0, 0.0, 0.0 };
  
  /**
   * La constance <b>IDENTITY</b> reprsente le tableau d'une matrice identit.
   */
  private static final double[] IDENTITY = { 1.0, 0.0, 0.0, 0.0,
                                             0.0, 1.0, 0.0, 0.0,
                                             0.0, 0.0, 1.0, 0.0,
                                             0.0, 0.0, 0.0, 1.0 };
  
  //-------------
  // VARIABLES //
  //-------------
  
	/**
	 * La variable 'matrix' correspond  la matrice 4x4 contenant les 16 lments.
	 */
	private final double[] matrix;
	
	//-----------------
	// CONSTRUCTEURS //
	//-----------------
	
	/**
	 * Constructeur de la matrice nulle (tous les lments M[i][j] seront nuls.)  
	 */
	public SMatrix4x4()
	{
		this(EMPTY);
	}

	/**
	 * Constructeur d'une matrice 4x4 avec paramtres en coordonne M[i][j] tel que i correspond  la ligne
	 * et j correspond  la colonne. Visuellement, l'organisation des donnes correspond  la matrice suivante :
	 * <p>M = | m00  m01  m02  m03 |</p>
	 * <ul>| m10  m11  m12  m13 |</ul>
	 * <ul>| m20  m21  m22  m23 |</ul>
	 * <ul>| m30  m31  m32  m33 |</ul>
	 * 
	 * @param m00 - L'lment 00 de la matrice (colonne 1 : ligne 1)
	 * @param m01 - L'lment 01 de la matrice (colonne 2 : ligne 1)
	 * @param m02 - L'lment 02 de la matrice (colonne 3 : ligne 1)
	 * @param m03 - L'lment 03 de la matrice (colonne 4 : ligne 1)
	 * @param m10 - L'lment 10 de la matrice (colonne 1 : ligne 2)
   * @param m11 - L'lment 11 de la matrice (colonne 2 : ligne 2)
   * @param m12 - L'lment 12 de la matrice (colonne 3 : ligne 2)
   * @param m13 - L'lment 13 de la matrice (colonne 4 : ligne 2)
   * @param m20 - L'lment 20 de la matrice (colonne 1 : ligne 3)
   * @param m21 - L'lment 21 de la matrice (colonne 2 : ligne 3)
   * @param m22 - L'lment 22 de la matrice (colonne 3 : ligne 3)
   * @param m23 - L'lment 23 de la matrice (colonne 4 : ligne 3)
   * @param m30 - L'lment 30 de la matrice (colonne 1 : ligne 4)
   * @param m31 - L'lment 31 de la matrice (colonne 2 : ligne 4)
   * @param m32 - L'lment 32 de la matrice (colonne 3 : ligne 4)
   * @param m33 - L'lment 33 de la matrice (colonne 4 : ligne 4) 
	 */
	public SMatrix4x4(double m00, double m01, double m02, double m03,
        			      double m10, double m11, double m12, double m13,
        			      double m20, double m21, double m22, double m23,
        			      double m30, double m31, double m32, double m33)
	{
	  this.matrix = new double[]{ m00,	  m01,	  m02,	 m03,
	                              m10,	  m11,	  m12,	 m13,
	                              m20,	  m21,	  m22,	 m23,
	                              m30,	  m31,	  m32,	 m33  };
	}
	
	/**
	 * Constructeur d'une matrice avec tableau des 16 composantes de la matrice.
	 * 
	 * @param m Le tableau des 16 composantes de la matrice.
	 */
  private SMatrix4x4(double[] m)
  {
    this.matrix = m;
  }
	
  //------------
  // MTHODES //
  //------------
  
  /**
   * Mthode pour obtenir une composante de la matrice en indice i de la matrice o i correspond  la position d'un lment organis dans un tableau de valeur entre 0 et 15.
   *
   * @return La composante en position i dans le tableau des valeurs de la matrice.
   * @throws ArrayIndexOutOfBoundsException Si l'accs dans le tableau de la matrice n'est pas adquat.
   */
  public double get(int i) throws ArrayIndexOutOfBoundsException
  {
    return this.matrix[i];
  }
  
  /**
   * Mthode pour obtenir une valeur m[i,j] de la matrice. 
   * 
   * @param i L'indice de la ligne de la matrice.
   * @param j L'indice de la colonne de la matrice.
   * @return L'lment m[i,j] de la matrice.
   * @throws ArrayIndexOutOfBoundsException Si la coordonne (i,j) n'est pas  l'intrieure de (0,0)  (3,3) pour une matrice 4x4.
   */
  public double get(int i, int j) throws ArrayIndexOutOfBoundsException
  {
    if(i < 0 || j < 0)
      throw new ArrayIndexOutOfBoundsException("Erreur SMatrix4x4 001 : L'accs (i,j) = (" + i + "," + j + ") contient une valeur ngative.");
        
    return this.get(i*4 + j);
  }
  
  /**
   * Mthode pour obtenir une copie de la matrice en format tableau des donnes.
   * 
   * @return Un tableau correspondant  la matrice.
   */
  public double[] toArray()
  {
    return this.matrix.clone();
  }
  
  /**
   * ...
   * 
   * @return
   */
  public DoubleBuffer toDoubleBuffer()
  {
    return DoubleBuffer.wrap(this.matrix.clone());
  }
  
  /**
   * Mthode qui effecteur l'addition de deux matrices.
   * 
   * @param M La matrice  additionner.
   * @return La matrice correspondant  la somme des deux matrices.
   */
  public SMatrix4x4 add(SMatrix4x4 M)
  {
    return new SMatrix4x4(SArrays.add(this.matrix, M.matrix));
  }
  
  /**
   * Mthode qui effecteur la soustracteur de deux matrices.
   * 
   * @param M La matrice  soustraire.
   * @return La matrice correspondant  la soustraction des deux matrices.
   */
  public SMatrix4x4 substract(SMatrix4x4 M)
  {
    return new SMatrix4x4(SArrays.substract(this.matrix, M.matrix));
  }
  
  /**
   * Mthode qui effectue la multiplication de la matrice par un scalaire.
   * 
   * @param v La valeur  multiplier.
   * @return La matrice multiplie par par un scalaire.
   */
  public SMatrix4x4 multiply(double v)
  {
    return new SMatrix4x4(SArrays.multiply(this.matrix, v));
  }
  
	/**
	 * Mthode qui effectue le produit matriciel entre deux matrices.
	 * La <b>matrice M1</b> qui lance l'appel de cette mthode sera la matrice de <b>gauche</b> et la <b>matrice M2</b>
	 * passe en paramtre sera la matrice de <b>droite</b>. Cette prcision est ncessaire tant donn
	 * que le <b>produit matriciel n'est pas commutatif</b>.
	 * 
	 * <p>Le rsutat du produit matriciel M = M1*M2 correspond au calcul</p>
	 * <ul>M[i][j] = SOMME sur k : M1[i][k]*M2[k][j]</ul>
	 * <p>o M[i][j] est une composantes de la matrice M.</p>
	 * 
	 * @param M - La matrice M2 de droite  multiplier avec la matrice M1 lanant l'appel de la mthode.
	 * @return Le rsultat M du produit matriciel.
	 */
	public SMatrix4x4 multiply(SMatrix4x4 M) 
	{
		double[] m = M.matrix;
		
		return new SMatrix4x4(	
				
				// Premiere ligne
				matrix[0]*m[0] + matrix[1]*m[4] + matrix[2]*m[8] + matrix[3]*m[12],         
				matrix[0]*m[1] + matrix[1]*m[5] + matrix[2]*m[9] + matrix[3]*m[13],
				matrix[0]*m[2] + matrix[1]*m[6] + matrix[2]*m[10] + matrix[3]*m[14],
				matrix[0]*m[3] + matrix[1]*m[7] + matrix[2]*m[11] + matrix[3]*m[15],

				// Deuxieme ligne
				matrix[4]*m[0] + matrix[5]*m[4] + matrix[6]*m[8] + matrix[7]*m[12],         
				matrix[4]*m[1] + matrix[5]*m[5] + matrix[6]*m[9] + matrix[7]*m[13],
				matrix[4]*m[2] + matrix[5]*m[6] + matrix[6]*m[10] + matrix[7]*m[14],
				matrix[4]*m[3] + matrix[5]*m[7] + matrix[6]*m[11] + matrix[7]*m[15],

				// Troisieme ligne
				matrix[8]*m[0]	+ matrix[9]*m[4] + matrix[10]*m[8] + matrix[11]*m[12],        
				matrix[8]*m[1]	+ matrix[9]*m[5] + matrix[10]*m[9] + matrix[11]*m[13],
				matrix[8]*m[2]	+ matrix[9]*m[6] + matrix[10]*m[10] + matrix[11]*m[14],
				matrix[8]*m[3]	+ matrix[9]*m[7] + matrix[10]*m[11] + matrix[11]*m[15],

				// Quatrieme ligne
				matrix[12]*m[0]	+ matrix[13]*m[4] + matrix[14]*m[8]	+ matrix[15]*m[12], 				
				matrix[12]*m[1]	+ matrix[13]*m[5] + matrix[14]*m[9]	+ matrix[15]*m[13],
				matrix[12]*m[2]	+ matrix[13]*m[6] + matrix[14]*m[10] + matrix[15]*m[14],
				matrix[12]*m[3]	+ matrix[13]*m[7] + matrix[14]*m[11] + matrix[15]*m[15]
						
				);
	}

	/**
	 * Mthode qui effectue le produit entre une matrice 4x4 et un vecteur colonne en 4d.
	 * La <b>matrice M</b> qui lance l'appel de cette mthode sera  <b>gauche</b> et le <b>vecteur colonne</b> pass
	 * en paramtre sera  <b>droite</b>. Cette prcision est ncessaire tant donn
	 * que le <b>produit entre une matrice et un vecteur n'est pas commutatif</b>.
	 * 
	 * <p>Le rsutat du produit matriciel u = M*v correspond au calcul</p>
	 * <ul>u[i] = SOMME sur k : M[i][k]*v[k]</ul>
	 * <p>o u[i] est une composantes du vecteur colonne u.</p>
	 * 
	 * @param v - Le vecteur de droite  multiplier avec la matrice qui lance l'appel de la mthode.
	 * @return Le vecteur colonne u tant le rsultat du produit de la matrice M avec le vecteur colonne v.
	 */
	public SVector4d multiply(SVector4d v)
	{
		return new SVector4d(	
				
				matrix[0]*v.getX() + matrix[1]*v.getY() + matrix[2]*v.getZ() + matrix[3]*v.getT(),
				matrix[4]*v.getX() + matrix[5]*v.getY() + matrix[6]*v.getZ() + matrix[7]*v.getT(),
				matrix[8]*v.getX() + matrix[9]*v.getY() + matrix[10]*v.getZ() + matrix[11]*v.getT(),
				matrix[12]*v.getX() + matrix[13]*v.getY() + matrix[14]*v.getZ() + matrix[15]*v.getT()
				
				);
	}
	
	/**
	 * Mthode qui effectue le produit entre une matrice 4x4 et un vecteur colonne en 3d dont une 4ime dimension t = 1.0 est ajoute.
	 * La <b>matrice M</b> qui lance l'appel de cette mthode sera  <b>gauche</b> et le <b>vecteur colonne</b> pass
	 * en paramtre sera  <b>droite</b>. Cette prcision est ncessaire tant donn
	 * que le <b>produit entre une matrice et un vecteur n'est pas commutatif</b>.
	 * 
	 * <p>Le rsutat du produit matriciel u = M*v correspond au calcul</p>
	 * <ul>u[i] = SOMME sur k : M[i][k]*v[k]</ul>
	 * <p>o u[i] est une composantes du vecteur colonne u.</p>
	 * 
	 * @param v - Le vecteur de droite  multiplier avec la matrice qui lance l'appel de la mthode.
	 * @return Le vecteur colonne u tant le rsultat du produit de la matrice M avec le vecteur colonne v.
	 */
  public SVector4d multiply(SVector3d v)
  {
    return multiply(new SVector4d(v));
  }
  
	/**
	 * <p>Mthode pour obtenir la matrice identit I.</p> 
	 * <p>Soit une matrice identit I et une
	 * matrice A quelconque, alors</p> 
	 * 
	 *  <ul> A = A*I = I*A </ul>
	 *  <p>ce qui rend la <b>matrice identit commutative</b>.</p>
	 * 
	 * @return La matrice identit I.
	 */
	public static SMatrix4x4 identity()
	{
		return new SMatrix4x4(IDENTITY);
	}
	
	/**
	 * Mthode pour obtenir la matrice transpose de la matrice qui lance l'appel de cette mthode.
	 * La matrice transpose B s'obtient  partir d'une matrice A grce au calcul
   *
   * <ul> B[j][i] = A[i][j]</ul>
   * 
	 * @return La matrice transpose.
	 */
	public SMatrix4x4 transpose()
	{
		return new SMatrix4x4(
		    
		    this.matrix[0],  this.matrix[4],  this.matrix[8],  this.matrix[12],
		    this.matrix[1],  this.matrix[5],  this.matrix[9],  this.matrix[13],
		    this.matrix[2],  this.matrix[6],  this.matrix[10], this.matrix[14],
		    this.matrix[3],  this.matrix[7],  this.matrix[11], this.matrix[15]
      
		    );
	}
	
	/**
	 * Mthode raliser l'inversion de la matrice. L'inversion d'une matrice est possible uniquement si son dterminant n'est pas nul (det != 0).
	 * 
	 * Rfrence : https://semath.info/src/inverse-cofactor-ex4.html
	 * 
	 * @return La matrice inverse.
	 * @throws IllegalArgumentException Si la matrice ne peut pas tre inverse (dterminant nul).
	 */
	public SMatrix4x4 inverse() throws IllegalArgumentException
	{
	  // valuer le dterminant de la matrice.
	  double det = this.determinant();
	  
	  // Si le dterminant est zro, c'est que la matrice ne peut pas tre inverse.
	  if(SMath.nearlyZero(det))
	    throw new IllegalArgumentException("Erreur SMatrix4x4 002 : Le dterminant de la matrice tant nul, la matrice ne peut pas tre inverse.");
	  
	  double a11 = this.matrix[0];
    double a12 = this.matrix[1];
    double a13 = this.matrix[2];
    double a14 = this.matrix[3];
    
    double a21 = this.matrix[4];
    double a22 = this.matrix[5];
    double a23 = this.matrix[6];
    double a24 = this.matrix[7];
    
    double a31 = this.matrix[8];
    double a32 = this.matrix[9];
    double a33 = this.matrix[10];
    double a34 = this.matrix[11];
    
    double a41 = this.matrix[12];
    double a42 = this.matrix[13];
    double a43 = this.matrix[14];
    double a44 = this.matrix[15];
    
    double A11 = a22*a33*a44 + a23*a34*a42 + a24*a32*a43 - a24*a33*a42 - a23*a32*a44 - a22*a34*a43;
    double A12 = -a12*a33*a44 - a13*a34*a42 - a14*a32*a43 + a14*a33*a42 + a13*a32*a44 + a12*a34*a43;
    double A13 = a12*a23*a44 + a13*a24*a42 + a14*a22*a43 - a14*a23*a42 - a13*a22*a44 - a12*a24*a43;
    double A14 = -a12*a23*a34 - a13*a24*a32 - a14*a22*a33 + a14*a23*a32 + a13*a22*a34 + a12*a24*a33;
    
    double A21 = -a21*a33*a44 - a23*a34*a41 - a24*a31*a43 + a24*a33*a41 + a23*a31*a44 + a21*a34*a43;
    double A22 = a11*a33*a44 + a13*a34*a41 + a14*a31*a43 - a14*a33*a41  -a13*a31*a44 - a11*a34*a43;
    double A23 = -a11*a23*a44 - a13*a24*a41 - a14*a21*a43 + a14*a23*a41 + a13*a21*a44 + a11*a24*a43;
    double A24 = a11*a23*a34 + a13*a24*a31 + a14*a21*a33 - a14*a23*a31 - a13*a21*a34 - a11*a24*a33;
    
    double A31 = a21*a32*a44 + a22*a34*a41 + a24*a31*a42 - a24*a32*a41 - a22*a31*a44 - a21*a34*a42;
    double A32 = -a11*a32*a44 - a12*a34*a41 - a14*a31*a42 + a14*a32*a41 + a12*a31*a44 + a11*a34*a42;
    double A33 = a11*a22*a44 + a12*a24*a41 + a14*a21*a42 - a14*a22*a41 - a12*a21*a44 - a11*a24*a42;
    double A34 = -a11*a22*a34 - a12*a24*a31 - a14*a21*a32 + a14*a22*a31 + a12*a21*a34 + a11*a24*a32;
    
    double A41 = -a21*a32*a43 - a22*a33*a41 - a23*a31*a42 + a23*a32*a41 + a22*a31*a43 + a21*a33*a42;
    double A42 = a11*a32*a43 + a12*a33*a41 + a13*a31*a42 - a13*a32*a41 - a12*a31*a43 - a11*a33*a42;
    double A43 = -a11*a22*a43 - a12*a23*a41 - a13*a21*a42 + a13*a22*a41 + a12*a21*a43 + a11*a23*a42;
    double A44 = a11*a22*a33 + a12*a23*a31 + a13*a21*a32 - a13*a22*a31 - a12*a21*a33 - a11*a23*a32;
    
    return new SMatrix4x4(A11/det, A12/det, A13/det, A14/det,
                          A21/det, A22/det, A23/det, A24/det,
                          A31/det, A32/det, A33/det, A34/det,
                          A41/det, A42/det, A43/det, A44/det);
        
	}
	
	/**
	 * Mthode pour obtenir le dterminant de la matrice.
	 * 
	 * Rfrence : https://semath.info/src/inverse-cofactor-ex4.html
	 * 
	 * @return Le dterminant de la matrice.
	 */
	public double determinant()
	{
	  double a11 = this.matrix[0];
	  double a12 = this.matrix[1];
	  double a13 = this.matrix[2];
	  double a14 = this.matrix[3];
	  
	  double a21 = this.matrix[4];
    double a22 = this.matrix[5];
    double a23 = this.matrix[6];
    double a24 = this.matrix[7];
    
    double a31 = this.matrix[8];
    double a32 = this.matrix[9];
    double a33 = this.matrix[10];
    double a34 = this.matrix[11];
    
    double a41 = this.matrix[12];
    double a42 = this.matrix[13];
    double a43 = this.matrix[14];
    double a44 = this.matrix[15];
    
    double t1 = a11*(a22*a33*a44 + a23*a34*a42 + a24*a32*a43 - a24*a33*a42 - a23*a32*a44 - a22*a34*a43);
    double t2 = -a21*(a12*a33*a44 + a13*a34*a42 + a14*a32*a43 - a14*a33*a42 - a13*a32*a44 - a12*a34*a43);
    double t3 = a31*(a12*a23*a44 + a13*a24*a42 + a14*a22*a43 - a14*a23*a42 - a13*a22*a44 - a12*a24*a43);
    double t4 = -a41*(a12*a23*a34 + a13*a24*a32 + a14*a22*a33 - a14*a23*a32 - a13*a22*a34 - a12*a24*a33);
    
    return t1 + t2 + t3 + t4;
	}
	
	/**
	 * Mthode pour obtenir une matrice de transformation linaire de translation Tr <b>par rapport
	 *  l'origine</b> d'un systme d'axe cartsien <i>xyz</i> si la matrice est utilise dans l'ordre
	 * <ul>u = Tr * v</ul>
	 * <p>o v est le vecteur  transformer et u est le vecteur transform.</p>
	 * 
	 * @param v - Le vecteur de translation en coordonne cartsienne.
	 * @return La matrice de translation T.
	 */
	public static SMatrix4x4 translation(SVector3d v)
	{
		return SMatrix4x4.translation(v.getX(), v.getY(), v.getZ());
	}
	
	/**
	 * Mthode pour obtenir une matrice de transformation linaire de translation Tr <b>par rapport
	 *  l'origine</b> d'un systme d'axe cartsien <i>xyz</i> si la matrice est utilise dans l'ordre
	 * <ul>u = Tr * v</ul>
	 * <p>o v est le vecteur  transformer et u est le vecteur transform.</p>
	 * 
	 * @param tr_x - La translation selon l'axe <i>x</i>.
	 * @param tr_y - La translation selon l'axe <i>y</i>.
	 * @param tr_z - La translation selon l'axe <i>z</i>.
	 * @return La matrice de translation T.
	 */
	public static SMatrix4x4 translation(double tr_x, double tr_y, double tr_z)
	{
		return new SMatrix4x4(
				
				1, 0, 0, tr_x,
			  0, 1, 0, tr_y,
			  0, 0, 1, tr_z,
			  0, 0, 0,  1
	  
				);
	}
	
	/**
	 * Mthode pour obtenir une matrice de transformation linaire d'homothtie Sc (scale) <b>par rapport
	 *  l'origine</b> d'un systme d'axe cartsien <i>xyz</i> si la matrice est utilise dans l'ordre
	 * <ul>u = Sc * v</ul>
	 * <p>o v est le vecteur  transformer et u est le vecteur transform.</p>
	 * 
	 * @param v - Le vecteur d'homothtie en coordonne cartsienne.
	 * @return La matrice d'homothtie Sc (scale).
	 */
	public static SMatrix4x4 scale(SVector3d v)
	{ 
		return SMatrix4x4.scale(v.getX(), v.getY(), v.getZ());
	}

	/**
	 * Mthode pour obtenir une matrice de transformation linaire d'homothtie Sc (scale) <b>par rapport
	 *  l'origine</b> d'un systme d'axe cartsien <i>xyz</i> si la matrice est utilise dans l'ordre
	 * <ul>u = Sc * v</ul>
	 * <p>o v est le vecteur  transformer et u est le vecteur transform.</p>
	 * 
	 * @param sx - L'homothtie selon l'axe <i>x</i>.
	 * @param sy - L'homothtie selon l'axe <i>y</i>.
	 * @param sz - L'homothtie selon l'axe <i>z</i>.
	 * @return La matrice de transformation d'homothtie Sc (scale).
	 */
	public static SMatrix4x4 scale(double sc_x,double sc_y,double sc_z)
	{
	  return new SMatrix4x4(	
			  
			  sc_x,    0,    0,   0,
			     0, sc_y,    0,   0,
			     0,    0, sc_z,   0,
			     0,    0,    0,   1
			   
			  );
	}

	/**
	 * Mthode pour obtenir une matrice de transformation linaire de rotation Rx autour de l'axe <i>x</i> 
	 * <b>par rapport  l'origine</b> d'un systme d'axe cartsien <i>xyz</i> si la matrice est utilise dans l'ordre
	 * <ul>u = Rx * v</ul>
	 * <p>o v est le vecteur  transformer et u est le vecteur transform.</p>  
	 * 
	 * @param degree - L'angle de rotation en degr.
	 * @return La matrice de rotation Rx autour de l'axe <i>x</i>.
	 */
	public static SMatrix4x4 rotationX(double degree)
	{
	  double angle = Math.toRadians(degree);
	  double sinO = Math.sin(angle);
	  double cosO = Math.cos(angle);

	  return new SMatrix4x4(	
			  
			  1,    0,     0, 0,
			  0, cosO, -sinO, 0,
			  0, sinO,  cosO, 0,
			  0,    0,     0, 1
			  
			  );
	}
	
	/**
	 * Mthode pour obtenir une matrice de transformation linaire de rotation Ry autour de l'axe <i>y</i> 
	 * <b>par rapport  l'origine</b> d'un systme d'axe cartsien <i>xyz</i> si la matrice est utilise dans l'ordre
	 * <ul>u = Ry * v</ul>
	 * <p>o v est le vecteur  transformer et u est le vecteur transform.</p>  
	 * 
	 * @param degree - L'angle de rotation en degr.
	 * @return La matrice de rotation Ry autour de l'axe <i>y</i>.
	 */
	public static SMatrix4x4 rotationY(double degree)
	{
	  double angle = Math.toRadians(degree);
	  double sinO = Math.sin(angle);
	  double cosO = Math.cos(angle);

	  return new SMatrix4x4(	
			  
          cosO,	  0, sinO,   0,
				     0, 	1,    0,   0,
				 -sinO, 	0, cosO,   0,
             0, 	0,    0,   1

	      );
	}
	
	/**
	 * Mthode pour obtenir une matrice de transformation linaire de rotation Rz autour de l'axe <i>z</i> 
	 * <b>par rapport  l'origine</b> d'un systme d'axe cartsien <i>xyz</i> si la matrice est utilise dans l'ordre
	 * <ul>u = Rz * v</ul>
	 * <p>o v est le vecteur  transformer et u est le vecteur transform.</p> 
	 *    
	 * @param degree - L'angle de rotation en degr.
	 * @return La matrice de rotation Rz autour de l'axe <i>z</i>.
	 */
	public static SMatrix4x4 rotationZ(double degree)
	{
	  double angle = Math.toRadians(degree);
	  double sinO = Math.sin(angle);
	  double cosO = Math.cos(angle);

	  return new SMatrix4x4(	
			 
			  cosO, -sinO,   0,   0,
			  sinO,  cosO,   0,   0,
			     0,     0,   1,   0,
			     0,     0,   0,   1

			  );
	}
	
	/**
   * <p>
   * Mthode pour obtenir une matrice de transformation linaire de rotation Ru autour d'un axe <i>u</i> quelconque
   * <b>par rapport  l'origine</b> d'un systme d'axe cartsien <i>xyz</i> si la matrice est utilise dans l'ordre
   * <ul>v1 = Ru * v0</ul>
   * o v0 est le vecteur  transformer et v1 est le vecteur transform. Dans la littrature, cette matrice porte le
   * nom de <b>matrice des cosinus directionnels</b> (<b><i>direction cosine matrix</i></b>). 
   * </p>
   * 
   * @param u L'axe de rotation (doit tre unitaire). 
   * @param angle L'angle de rotation en degr.
   * @return La matrice de rotation autour de l'axe <i>u</i>.
   * @throws SImpossibleNormalizationException Si l'axe de rotation <i>u</i> ne peut pas tre normalis.
   */
  public static SMatrix4x4 rotationU(SVector3d u, double angle) throws SImpossibleNormalizationException
  {
    // Vrification de la normalisation.
    u = u.normalize();
    
    double ux = u.getX();
    double uy = u.getY();
    double uz = u.getZ();
    
    double angle_u = Math.toRadians(angle);
    
    double cu = Math.cos(angle_u);
    double su = Math.sin(angle_u);
    
    // Reprsentation de la matrice de rotation autour de l'axe u
    double[] m = new double[]{
        
        ux*ux + (1 - ux*ux)*cu, ux*uy*(1 - cu) - uz*su, ux*uz*(1 - cu) + uy*su, 0.0,
        ux*uy*(1 - cu) + uz*su, uy*uy + (1 - uy*uy)*cu, uy*uz*(1 - cu) - ux*su, 0.0,
        ux*uz*(1 - cu) - uy*su, uy*uz*(1 - cu) + ux*su, uz*uz + (1 - uz*uz)*cu, 0.0,
                           0.0,                    0.0,                    0.0, 1.0 };
    
    return new SMatrix4x4(m);
  }
  
	/**
	 * <p>
	 * Mthode pour obtenir une matrice de transformation linaire de rotation Rzyx autour des 
	 * axes successifs <i>x</i>, <i>y</i> et <i>z</i> <b>par rapport  l'origine</b> 
	 * d'un systme d'axe cartsien <i>xyz</i> si la matrice est utilise dans l'ordre
	 * <ul>u = Rzyx * v</ul>
	 * <p>
	 * o v est le vecteur  transformer et u est le vecteur transform.
	 * </p> 
	 * <p>
	 * La matrice Rzyx est le rsultat de l'opration
	 * </p>
	 * <ul> Rzyx = Rz*Ry*Rx</ul>
	 * <p>
	 * o Rx est la matrice de rotation autour de l'axe <i>x</i>, 
	 *    Ry est la matrice de rotation autour de l'axe <i>y</i> et
	 *    Rz est la matrice de rotation autour de l'axe <i>z</i>.
	 * </p>
	 * </p>
	 * @param v Le vecteur de rotation selon l'axe x,y et z en <b>degr</b>.
	 * @return La matrice de rotation Rzyx dans l'ordre <i>x</i>, <i>y</i> et <i>z</i> si elle est utilise tel que u = Rzyx*v.
	 */
	public static SMatrix4x4 Rzyx(SVector3d v)
	{
	  //---------------------------------------------------------------------------------------------
	  //POUR LABO :
    //throw new SNoImplementationException("Erreur SMatrix4x4 : La mthode n'est pas implmente.");
	  //---------------------------------------------------------------------------------------------
	  
	  double angleX = Math.toRadians(v.getX());
		double angleY = Math.toRadians(v.getY());
		double angleZ = Math.toRadians(v.getZ());
		
		double cx = Math.cos(angleX);
		double sx = Math.sin(angleX);
		
		double cy = Math.cos(angleY);
		double sy = Math.sin(angleY);
		
		double cz = Math.cos(angleZ);
		double sz = Math.sin(angleZ);

    //Rsultat algbrique de la multiplication Rzyx = Rz * Ry * Rx
		double[] m = new double[]{
		    
		    cy*cz,  sx*sy*cz - cx*sz, cx*sy*cz + sx*sz,   0,
        cy*sz,  sx*sy*sz + cx*cz, cx*sy*sz - sx*cz,   0,
        -1*sy,             sx*cy,            cx*cy,   0,
            0,                 0,                0,   1  };
		
    return new SMatrix4x4(m);
	}
	
	/**
   * <p>
   * Mthode pour obtenir une matrice de transformation linaire de rotation Rxyz autour des 
   * axes successifs <i>z</i>, <i>y</i> et <i>x</i> <b>par rapport  l'origine</b> 
   * d'un systme d'axe cartsien <i>xyz</i> si la matrice est utilise dans l'ordre
   * <ul>u = Rxyz * v</ul>
   * <p>
   * o v est le vecteur  transformer et u est le vecteur transform.
   * </p> 
   * <p>
   * La matrice Rxyz est le rsultat de l'opration
   * </p>
   * <ul> Rxyz = Rx*Ry*Rz</ul>
   * <p>
   * o Rx est la matrice de rotation autour de l'axe <i>x</i>, 
   *    Ry est la matrice de rotation autour de l'axe <i>y</i> et
   *    Rz est la matrice de rotation autour de l'axe <i>z</i>.
   * </p>
   * </p>
   * @param v Le vecteur de rotation selon l'axe x,y et z en <b>degr</b>.
   * @return La matrice de rotation Rxyz dans l'ordre <i>z</i>, <i>y</i> et <i>x</i> si elle est utilise tel que u = Rxyz*v.
   */
  public static SMatrix4x4 Rxyz(SVector3d v)
  {
    //---------------------------------------------------------------------------------------------
    //POUR LABO :
    //throw new SNoImplementationException("Erreur SMatrix4x4 : La mthode n'est pas implmente.");
    //---------------------------------------------------------------------------------------------
    
    double angleX = Math.toRadians(v.getX());
    double angleY = Math.toRadians(v.getY());
    double angleZ = Math.toRadians(v.getZ());
    
    double cx = Math.cos(angleX);
    double sx = Math.sin(angleX);
    
    double cy = Math.cos(angleY);
    double sy = Math.sin(angleY);
    
    double cz = Math.cos(angleZ);
    double sz = Math.sin(angleZ);

    //Rsultat algbrique de la multiplication Rxyz = Rx * Ry * Rz
    double[] m = new double[]{
        
                   cy*cz,         -1*cy*sz,       sy,   0,
        cx*sz + sx*sy*cz, cx*cz - sx*sy*sz, -1*sx*cy,   0,
        sx*sz - cx*sy*cz, sx*cz + cx*sy*sz,    cx*cy,   0,
                       0,                0,        0,   1  };
        
    return new SMatrix4x4(m);
  }
  
	/**
	 * <p>Mthode pour obtenir une matrice de transformation linaire d'homothtie Sc (<i>scale</i>), de rotation Rzyx autour des 
   * axes successifs <i>x</i>, <i>y</i> et <i>z</i> puis de translation Tr <b>par rapport  l'origine</b> 
   * d'un systme d'axe cartsien <i>xyz</i>. La matrice TRzyxS est le rsultat de l'opration
   * <ul> TrRzyxSc = Tr*Rz*Ry*Rx*Sc</ul>
   * et elle doit tre utilise sous la reprsentation
   * <ul> u = TrRzyxSc*v </ul>
   * o v est le vecteur  transformer et u est le vecteur transform.</p>
	 * @param translation - Le vecteur reprsentant la translation selon l'axe <i>x</i>, <i>y</i> et <i>z</i>.
	 * @param rotation - Le vecteur rotation autour des avec <i>x</i>, <i>y</i> et <i>z</i> successivement.
	 * @param scale - Le vecteur d'homothtie (<i>scale</i>)  appliquer.
	 * @return La matrice de transformation linaire TrRzyxSc.
	 */
	public static SMatrix4x4 TrRzyxSc(SVector3d translation, SVector3d rotation, SVector3d scale)
	{
	  //---------------------------------------------------------------------------------------------
	  //POUR LABO :
    //throw new SNoImplementationException("Erreur SMatrix4x4 : La mthode n'est pas implmente.");
	  //---------------------------------------------------------------------------------------------
	  
	  double angleX = Math.toRadians(rotation.getX());
    double angleY = Math.toRadians(rotation.getY());
    double angleZ = Math.toRadians(rotation.getZ());
    
    double cx = Math.cos(angleX);
    double sx = Math.sin(angleX);
    
    double cy = Math.cos(angleY);
    double sy = Math.sin(angleY);
    
    double cz = Math.cos(angleZ);
    double sz = Math.sin(angleZ);

    //Rsultat algbrique de la multiplication TrRzyxSc = Tr * Rz * Ry * Rx * Sc
    double[] m = new double[]{
        
        scale.getX()*cy*cz,  scale.getY()*(sx*sy*cz - cx*sz), scale.getZ()*(cx*sy*cz + sx*sz),   translation.getX(),
        scale.getX()*cy*sz,  scale.getY()*(sx*sy*sz + cx*cz), scale.getZ()*(cx*sy*sz - sx*cz),   translation.getY(),
        -1*scale.getX()*sy,               scale.getY()*sx*cy,              scale.getZ()*cx*cy,   translation.getZ(),
                         0,                                0,                               0,                    1  };
    
    return new SMatrix4x4(m);
	}
	
	/**
   * <p>
   * Mthode pour obtenir une matrice de transformation linaire de translation Tr, de rotation Rxyz autour des 
   * axes successifs <i>z</i>, <i>y</i> et <i>x</i> puis d'homothtie Sc (<i>scale</i>) <b>par rapport  l'origine</b> 
   * d'un systme d'axe cartsien <i>xyz</i>. La matrice ScRxyzTr est le rsultat de l'opration
   * <ul> ScRxyzTr = Sc*Rx*Ry*Rz*Tr</ul>
   * et elle doit tre utilise sous la reprsentation
   * <ul> u = ScRxyzTr*v .</ul>
   * o v est le vecteur  transformer et u est le vecteur transform.
   * </p>
   * @param scale - Le vecteur d'homothtie (<i>scale</i>)  appliquer.
   * @param rotation - Le vecteur rotation autour des avec <i>x</i>, <i>y</i> et <i>z</i> successivement.
   * @param translation - Le vecteur reprsentant la translation selon l'axe <i>x</i>, <i>y</i> et <i>z</i>.
   * @return La matrice de transformation linaire ScRxyzTr.
   */
  public static SMatrix4x4 ScRxyzTr(SVector3d scale, SVector3d rotation, SVector3d translation)
  {
    //---------------------------------------------------------------------------------------------
    //POUR LABO :
    //throw new SNoImplementationException("Erreur SMatrix4x4 : La mthode n'est pas implmente.");
    //---------------------------------------------------------------------------------------------
    
    double angleX = Math.toRadians(rotation.getX());
    double angleY = Math.toRadians(rotation.getY());
    double angleZ = Math.toRadians(rotation.getZ());
    
    double cx = Math.cos(angleX);
    double sx = Math.sin(angleX);
    
    double cy = Math.cos(angleY);
    double sy = Math.sin(angleY);
    
    double cz = Math.cos(angleZ);
    double sz = Math.sin(angleZ);

    double R10 = cx*sz + sx*sy*cz;
    double R11 = cx*cz - sx*sy*sz;
    double R20 = sx*sz - cx*sy*cz;
    double R21 = sx*cz + cx*sy*sz;
    
    double tr_x = translation.getX();
    double tr_y = translation.getY();
    double tr_z = translation.getZ();
    
    // Rsultat algbrique de la multiplication ScRxyzTr = Sc * Rx * Ry * Rz * Tr
    double[] m = new double[]{
        
        scale.getX()*(cy*cz), -1*scale.getX()*(cy*sz),       scale.getX()*(sy), scale.getX()*(cy*cz*tr_x - cy*sz*tr_y + sy*tr_z),
            scale.getY()*R10,        scale.getY()*R11, -1*scale.getY()*(sx*cy),  scale.getY()*(R10*tr_x + R11*tr_y - sx*cy*tr_z),
            scale.getZ()*R20,        scale.getZ()*R21,      scale.getZ()*cx*cy,  scale.getZ()*(R20*tr_x + R21*tr_y + cx*cy*tr_z),
                           0,                       0,                       0,                                                1  };
    
    return new SMatrix4x4(m);
  }
	
  /**
   * Mthode pour gnrer une matrice de rotation qui tourne un axe z dans l'orientation d'un vecteur.
   * 
   * @param v L'orientation vers lequel l'axe z sera tourn.
   * @return La matrice de rotation.
   * @throws IllegalArgumentExeception Si le vecteur v ne peut pas correspondre  une orientation.
   */
  public static SMatrix4x4 rotationZtoVector(SVector3d v) throws SImpossibleNormalizationException
  {   
    v = v.normalize();
    
    double thetaX;
    double thetaY;
    double thetaZ;
    
    double sigmaX;
    double sigmaY;
    
    double d = Math.sqrt(v.getY()*v.getY() + v.getZ()*v.getZ());
    
    // Cas gnral :
    if(!SMath.nearlyZero(d))
    {
      sigmaY = Math.asin(v.getY() / d);
      sigmaX = Math.atan(v.getX() / d);
      
      // Dterminer Z --> Valeur impose.
      thetaZ = Math.PI/2;
      
      // Dterminer X.
      if(SMath.nearlyEquals(v.getX(), Math.sin(sigmaX)))
        thetaX = sigmaX;
      else
        if(SMath.nearlyEquals(v.getX(), Math.sin(sigmaX + Math.PI)))
          thetaX = sigmaX + Math.PI;
        else
          throw new RuntimeException("Oh My God !!! en X");
      
      // Dterminer Y.
      if(SMath.nearlyEquals(v.getZ(), Math.cos(thetaX)*Math.cos(sigmaY)))
        thetaY = sigmaY;
      else
        if(SMath.nearlyEquals(v.getZ(), Math.cos(thetaX)*Math.cos(Math.PI - sigmaY)))
          thetaY = Math.PI - sigmaY;
        else
          throw new RuntimeException("Oh My God !!! en Y");
    }
    else
    {
      // Cas particulier : y et z sont gal  zro.
      
      // Le vecteur (0,0,1) doit tre tourn en (1,0,0) ou (-1,0,0).
      
      thetaX = 0.0;
      thetaY = Math.PI / 2;
      thetaZ = 0.0;
  
      // Si orient dans le sens positif de l'axe -x.
      if(v.getX() < 0.0)
        thetaY *= -1.0;
    }       
    
    return SMatrix4x4.Rzyx(new SVector3d(Math.toDegrees(thetaX), Math.toDegrees(thetaY), Math.toDegrees(thetaZ)));
    
  }
  
  @Override
  public String toString()
  {
    return  "\n"+
            "|" + matrix[0] + "\t" +  matrix[1] + "\t" +  matrix[2] + "\t" +  matrix[3] + "|"  + "\n" +  
            "|" + matrix[4] + "\t" +  matrix[5] + "\t" +  matrix[6] + "\t" +  matrix[7] + "|"  + "\n" +
            "|" + matrix[8] + "\t" +  matrix[9] + "\t" +  matrix[10] + "\t" +  matrix[11] + "|"  + "\n" +
            "|" + matrix[12] + "\t" +  matrix[13] + "\t" +  matrix[14] + "\t" +  matrix[15] + "|"+ "\n";
  }
	
	@Override
  public int hashCode() {
    final int prime = 31;
    int result = 1;
    result = prime * result + Arrays.hashCode(matrix);
    return result;
  }

  @Override
  public boolean equals(Object obj) {
    
    if (this == obj)
      return true;
    
    if (obj == null)
      return false;
    
    if (getClass() != obj.getClass())
      return false;
    
    SMatrix4x4 other = (SMatrix4x4) obj;
    
    if (!SArrays.nearlyEquals(matrix, other.matrix))
      return false;
    
    return true;
  }
	
}
