/**
 * 
 */
package sim.math;

import java.io.BufferedWriter;
import java.io.IOException;

import sim.exception.SConstructorException;

/**
 * La classe <b>SVectorPixel</b> reprsente un vecteur pour positionner un pixel dans un <b>viewport</b>.
 * 
 * @author Simon Vzina
 * @since 2014-12-27
 * @version 2017-12-14
 */
public class SVectorPixel implements SVector {

  //--------------
  // CONSTANTES //
  //--------------
  
  /**
   * La constante 'DEFAULT_COMPONENT' correspond  la coordonne par dfaut des composantes x et y du pixel qui est gale  {@value}. 
   */
  private final static int DEFAULT_COMPONENT = 0;
  
  //--------------
  // VARIABLES  //
  //--------------
  
	/**
	 * La variable 'x' correspond  la coordonne x du pixel.
	 */
  private final int x;		
  
  /**
   * La variable 'y' correspond  la coordonne y du pixel.
   */
	private final int y;		
	
	//-----------------------------
	// CONSTRUCTEURS ET MTHODES //
	//-----------------------------
	
	/**
	 * Constructeur d'un pixel par dfaut. Ce vecteur correspond  l'origine d'un viewport tant  la coordonne (0,0).
	 * Dans la grande majorit des viewports, cette coordonne correspond gomtriquement au coint suprieur gauche. 
	 */
	public SVectorPixel()
	{
		this(DEFAULT_COMPONENT, DEFAULT_COMPONENT);
	}
	
	/**
	 * Constructeur d'un vecteur de pixel.
	 * 
	 * @param x - La coordonne x du pixel.
	 * @param y - La coordonne y du pixel.
	 * @throws SConstructorException Si le pixel possde une coordonne ngative.
	 */
	public SVectorPixel(int x, int y) throws SConstructorException
	{
	  //Vrification des coordonnes afin qu'elles soient positives
	  if(x < 0)
	    throw new SConstructorException("Erreur SVectorPixel 001 : La coordonne x = " + x + " ne peut pas tre ngative.");
		
	  if(y < 0)
	    throw new SConstructorException("Erreur SVectorPixel 002 : La coordonne y = " + y + " ne peut pas tre ngative.");
			
	  this.x = x;
	  this.y = y;
	}

	/**
	 * Mthode pour obtenir la coordonne x du pixel.
	 * 
	 * @return La coordonne x du pixel.
	 */
	public int getX()
	{ 
	  return x;
	}
	
	/**
	 * Mthode pour obtenir la coordonne y du pixel.
	 * 
	 * @return La coordonne y du pixel.
	 */
	public int getY()
	{ 
	  return y;
	}
	
	@Override
	public SVector add(SVector v) 
  {
	  return add((SVectorPixel)v);
  }
  
	/**
	 * Mthode pour effectuer l'addition d'une coordonne de pixel avec une autre.
	 * 
	 * @param p - La coordonne de pixel  additionner.
	 * @return La coordonne de pixel rsultant de l'addition d'une autre coordonne de pixel.
	 */
	public SVectorPixel add(SVectorPixel p)
	{
	  return new SVectorPixel(x + p.x, y + p.y);
	}
	
	/**
	 * Mthode pour effectuer la multiplication par un scalaire avec une coordonne de pixel. Si la constante est ngative,
	 * une exception de type <b>SConstrutorException</b> sera lance puisqu'une coordonne de pixel ne peut pas contenir de composante ngative.
	 * 
	 * @param cst - La constante  multiplier avec la coordonne de pixel.
	 * @return La coordonne de pixel rsultant de la multiplication par un scalaire.
	 * @throws SConstructorException Si la constante  multiplier est ngative ne permettant pas de construire une coordonne de pixel avec des composantes ngatives.
	 */
	public SVectorPixel multiply(double cst) throws SConstructorException
	{
	  return new SVectorPixel((int)(x*cst), (int)(y*cst));
	}

	@Override
  public double dot(SVector v)
  {
    return dot((SVectorPixel)v);
  }
  
  /**
   * Mthode pour effectuer le <b>produit scalaire</b> entre deux vecteurs.
   * 
   * @param v Le vecteur  mettre en produit scalaire.
   * @return Le produit scalaire entre les deux vecteurs.
   */
  public double dot(SVectorPixel v)
  {
    return (x*v.x + y*v.y);
  }
  
	@Override
  public int hashCode()
  {
    final int prime = 31;
    int result = 1;
    result = prime * result + x;
    result = prime * result + y;
    return result;
  }

  @Override
  public boolean equals(Object obj)
  {
    if (this == obj) {
      return true;
    }
    
    if (obj == null) {
      return false;
    }
    
    if (!(obj instanceof SVectorPixel)) {
      return false;
    }
    
    SVectorPixel other = (SVectorPixel) obj;
    
    if (x != other.x) 
      return false;
    
    if (y != other.y) 
      return false;
    
    return true;
  }

  @Override
  public String toString()
  { 
    return new String("[" + x + ", " + y + "]");
  }
	
  @Override
  public void write(BufferedWriter bw) throws IOException
  {
    bw.write(toString());
  }
	
}//fin de la classe SVectorPixel
